import java.util.ArrayList;
import java.util.List;

import javafx.animation.FadeTransition;
import javafx.animation.Interpolator;
import javafx.animation.KeyFrame;
import javafx.animation.KeyValue;
import javafx.animation.Timeline;
import javafx.animation.TimelineBuilder;
import javafx.application.Application;
import javafx.beans.binding.Bindings;
import javafx.beans.binding.IntegerBinding;
import javafx.beans.property.DoubleProperty;
import javafx.beans.property.IntegerProperty;
import javafx.beans.property.SimpleDoubleProperty;
import javafx.beans.property.SimpleIntegerProperty;
import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.event.ActionEvent;
import javafx.event.EventHandler;
import javafx.scene.Group;
import javafx.scene.GroupBuilder;
import javafx.scene.Scene;
import javafx.scene.SceneBuilder;
import javafx.scene.control.Button;
import javafx.scene.control.Label;
import javafx.scene.effect.GaussianBlur;
import javafx.scene.input.MouseEvent;
import javafx.scene.paint.Color;
import javafx.scene.shape.Circle;
import javafx.scene.text.Font;
import javafx.stage.Stage;
import javafx.util.Duration;

public class Hunter extends Application {
	private static final int LEVEL_INCREASE = 5;
	private static final int START_LEVEL = 1;
	private static final int INITIAL_SCORE = 0;
	private static final int INITIAL_LIVES = 3;
	private final IntegerProperty score = new SimpleIntegerProperty(
			INITIAL_SCORE);
	private final IntegerProperty lives = new SimpleIntegerProperty(
			INITIAL_LIVES);
	private final IntegerBinding level = score.divide(LEVEL_INCREASE).add(
			START_LEVEL);
	private final List<Timeline> moveTimelines = new ArrayList<>();
	private Group enemyGroup;
	private Scene scene;
	private Timeline createEnemyTimeline;
	private Group particleGroup;

	public static void main(String[] args) {
		Application.launch(args);
	}

	@Override
	public void start(Stage primaryStage) throws Exception {
		primaryStage.setTitle("Hunter");

		level.addListener(new ChangeListener<Number>() {
			@Override
			public void changed(ObservableValue<? extends Number> observable,
					Number oldValue, Number newValue) {
				createEnemyTimeline.stop();
				startCreatingEnemies();
			}
		});

		enemyGroup = new Group();

		enemyGroup.effectProperty().bind(
				Bindings.when(lives.isEqualTo(0)).then(new GaussianBlur())
						.otherwise((GaussianBlur) null));

		particleGroup = new Group();

		Label information = new Label();
		information.setFont(new Font(24));
		information.textProperty().bind(
				Bindings.format("Score: %s, Level %s, Lives: %s", score, level,
						lives));

		Button restartButton = new Button("Restart");

		scene = SceneBuilder
				.create()
				.width(600)
				.height(400)
				.root(GroupBuilder
						.create()
						.children(information, enemyGroup, particleGroup,
								restartButton).build()).build();

		restartButton.translateXProperty().bind(
				scene.widthProperty().subtract(restartButton.widthProperty())
						.divide(2));

		restartButton.translateYProperty().bind(
				scene.heightProperty().subtract(restartButton.heightProperty())
						.divide(2));

		restartButton.visibleProperty().bind(lives.isEqualTo(0));

		restartButton.setOnAction(new EventHandler<ActionEvent>() {
			@Override
			public void handle(ActionEvent event) {
				restartGame();
			}
		});

		startCreatingEnemies();

		lives.addListener(new ChangeListener<Number>() {
			@Override
			public void changed(ObservableValue<? extends Number> observable,
					Number oldValue, Number newValue) {
				if (newValue.intValue() == 0) {
					stopGame();
				}
			}
		});

		primaryStage.setScene(scene);

		primaryStage.show();
	}

	private void startCreatingEnemies() {
		createEnemyTimeline = TimelineBuilder
				.create()
				.cycleCount(Timeline.INDEFINITE)
				.keyFrames(
						new KeyFrame(
								Duration.millis(100 + 1000d / level.get()),
								new EventHandler<ActionEvent>() {
									@Override
									public void handle(ActionEvent event) {
										createEnemy();
									}
								})).build();

		createEnemyTimeline.play();
	}

	private void createEnemy() {
		final Circle circle = new Circle(0, Math.random() * scene.getHeight(),
				20);

		circle.setOnMousePressed(new EventHandler<MouseEvent>() {
			@Override
			public void handle(MouseEvent event) {
				if (lives.get() > 0) {
					enemyGroup.getChildren().remove(circle);
					score.set(score.get() + 1);
					showExplosion(event.getSceneX(), event.getSceneY());
				}
			}
		});

		enemyGroup.getChildren().add(circle);

		Timeline move = new Timeline();
		move.getKeyFrames().add(
				new KeyFrame(Duration.seconds(3),
						new EventHandler<ActionEvent>() {
							@Override
							public void handle(ActionEvent event) {
								if (enemyGroup.getChildren().remove(circle)) {
									lives.set(lives.get() - 1);
								}
							}
						}, new KeyValue(circle.centerXProperty(), scene
								.getWidth(), Interpolator.EASE_BOTH)));
		move.play();

		moveTimelines.add(move);

	}

	private void showExplosion(double x, double y) {
		DoubleProperty dist = new SimpleDoubleProperty();

		final Group particles = new Group();

		for (int i = 0; i < 20; ++i) {
			Circle p = new Circle(3);
			p.setFill(Color.RED);
			double speedX = Math.random() - 0.5;
			double speedY = Math.random() - 0.5;
			p.centerXProperty().bind(dist.multiply(speedX).add(x));
			p.centerYProperty().bind(dist.multiply(speedY).add(y));
			particles.getChildren().add(p);
		}

		Timeline explode = new Timeline();
		explode.getKeyFrames().add(
				new KeyFrame(Duration.millis(500),
						new EventHandler<ActionEvent>() {
							@Override
							public void handle(ActionEvent event) {
								particleGroup.getChildren().remove(particles);
							}
						}, new KeyValue(dist, 200)));

		explode.play();

		FadeTransition fade = new FadeTransition();
		fade.setDuration(Duration.millis(500));
		fade.setNode(particles);
		fade.setFromValue(1);
		fade.setToValue(0);

		fade.play();

		particleGroup.getChildren().add(particles);

	}

	private void restartGame() {
		enemyGroup.getChildren().clear();
		lives.set(INITIAL_LIVES);
		score.set(INITIAL_SCORE);
		createEnemyTimeline.play();
	}

	private void stopGame() {
		createEnemyTimeline.stop();
		for (Timeline move : moveTimelines) {
			move.pause();
		}
		moveTimelines.clear();
	}
}
